@@ -23,7 +23,10 @@ int hvm_local_events_need_delivery(struct vcpu *v);
 static inline int local_events_need_delivery(void)
 {
     struct vcpu *v = current;
-    return (has_hvm_container_vcpu(v) ? hvm_local_events_need_delivery(v) :
+
+    ASSERT(!is_idle_vcpu(v));
+
+    return (is_hvm_vcpu(v) ? hvm_local_events_need_delivery(v) :
             (vcpu_info(v, evtchn_upcall_pending) &&
              !vcpu_info(v, evtchn_upcall_mask)));
 }
@@ -39,9 +42,15 @@ static inline void local_event_delivery_enable(void)
 }
 
 /* No arch specific virq definition now. Default to global. */
-static inline int arch_virq_is_global(uint32_t virq)
+static inline bool arch_virq_is_global(unsigned int virq)
 {
-    return 1;
+    return true;
 }
 
+#ifdef CONFIG_PV_SHIM
+# include <asm/pv/shim.h>
+# define arch_evtchn_is_special(chn) \
+             (pv_shim && (chn)->port && (chn)->state == ECS_RESERVED)
+#endif
+
 #endif
